//  Copyright (c) 1999-2016  Pavel Rychly, Milos Jakubicek

#include "config.hh"
#include "subcorp.hh"
#include "dynattr.hh"
#include "normattr.hh"
#include <finlib/binfile.hh>
#include <finlib/fromtof.hh>
#include <finlib/fsop.hh>
#include <cstdlib>
#include <fstream>
#include <limits>
#include <math.h>
#include <iostream>

using namespace std;

inline PosAttr *createDynAttr (CorpInfo::MSS &ao, const string &apath, 
                               const string &n, PosAttr *from,
                               bool ownedByCorpus = true)
{
    DynFun *fun = createDynFun (ao["FUNTYPE"].c_str(), 
                                ao["DYNLIB"].c_str(), ao["DYNAMIC"].c_str(),
                                ao["ARG1"].c_str(), ao["ARG2"].c_str());
    return createDynAttr (ao["DYNTYPE"], apath, n, fun, from, ao["LOCALE"],
                          CorpInfo::str2bool (ao["TRANSQUERY"]), ownedByCorpus);
}


PosAttr *Corpus::setup_attr (const string &attr_name)
{
    try {
        PosAttr* pa;
        string name, norm, path = conf->opts ["PATH"];
        size_t at = attr_name.find ('@');
        if (at != string::npos) {
            name = string (attr_name, 0, at);
            norm = string (attr_name, at + 1);
        } else
            name = attr_name;
        CorpInfo::MSS &ao = conf->find_attr (name);
        if (at != string::npos)
            pa = createNormPosAttr (get_attr (name), get_attr (norm));
        else if (ao["DYNAMIC"] != "")
            pa = createDynAttr (ao, path + name, name,
                                get_attr (ao["FROMATTR"]));
        else if (virt)
            pa = setup_virtposattr (virt, path + name, name, ao["LOCALE"],
                                    conf->opts["ENCODING"], true,
                                    ao["DEFAULTVALUE"], conf->opts["DOCSTRUCTURE"]);
        else
            pa = createPosAttr (ao["TYPE"], path + name, name, ao["LOCALE"], 
                                conf->opts["ENCODING"]);
        if (!conf->opts ["SUBCPATH"].empty()) {
            string spath = conf->opts ["SUBCPATH"];
            size_t slash = path.rfind('/');
            if (slash != string::npos && slash != path.length() - 1)
                spath += path.substr (slash + 1);
            pa = createSubCorpPosAttr (pa, spath, is_complement());
        }
        attrs.push_back (pair<string,PosAttr*> (name, pa));
        return pa;
    } catch (CorpInfoNotFound) {
        throw AttrNotFound (attr_name);
    }
}

Structure *Corpus::setup_struct (const string &name)
{
    try {
        CorpInfo *ci = conf->find_struct (name);
        Structure *s;
        if (virt) {
            s = new Structure (ci, name, virt);
        } else {
            string path = conf->opts ["PATH"];
            if (!conf->opts ["SUBCPATH"].empty())
                ci->opts ["SUBCPATH"] = conf->opts ["SUBCPATH"];
            s = new Structure (ci, path + name, name);
        }
        structs.push_back (pair<string,Structure*> (name, s));
        s->complement = is_complement();
        s->conf->conffile = conf->conffile;
        return s;
    } catch (CorpInfoNotFound) {
        throw AttrNotFound (name);
    }
}

void Corpus::init (CorpInfo *ci)
{
    hardcut = atol (ci->opts["HARDCUT"].c_str());
    maxctx = atol (ci->opts["MAXCONTEXT"].c_str());
    if (!ci->opts["ALIGNED"].empty()) {
        istringstream as (ci->opts["ALIGNED"]);
        string cname;
        while (getline (as, cname, ',')) {
            if (cname.empty())
                continue;
            aligned.push_back (AlignedCorpus (cname));
        }
    }
    if (ci->opts["VIRTUAL"] != "")
        virt = setup_virtcorp (ci->opts["VIRTUAL"]);
}

Corpus::Corpus (CorpInfo *ci, type_t t)
    : defaultattr (NULL), virt(NULL), type(t), conf (new CorpInfo (ci))
{
    init (conf);
}

Corpus::Corpus (const string &corp_name)
    : defaultattr (NULL), virt(NULL), type(Corpus_type),
      conf (loadCorpInfo (corp_name))
{
    init (conf);
    //if (!defaultattr)
    // XXX throw (no atributes)
}

Corpus::~Corpus ()
{
    for (VSA::iterator i = attrs.begin(); i != attrs.end(); i++)
        delete (*i).second;
    for (VSS::iterator i = structs.begin(); i != structs.end(); i++)
        delete (*i).second;
    delete conf;
    if (virt)
        delete virt;
    for (unsigned i = 0; i < aligned.size(); i++) {
        delete aligned[i].corp;
        delete_TokenLevel (aligned[i].level);
    }
}

PosAttr* Corpus::get_attr (const string &name, bool struct_attr)
{
    if (name == "-")
        return get_default_attr();

    VSA::iterator i = attrs.begin();
    while (i != attrs.end()) {
        if ((*i).first == name) 
            return (*i).second;
        i++;
    }

    int j;
    if ((j = name.find ('.')) >= 0) {
        string strucname (name, 0, j);
        string attrname (name, j+1);
        if (struct_attr) {
            Structure *s = get_struct (strucname);
            return s->get_attr (attrname);
        } else
            return get_struct_pos_attr (strucname, attrname);
    }
    return setup_attr (name);
}

PosAttr* Corpus::get_default_attr ()
{
    if (defaultattr)
        return defaultattr;
    else
        return defaultattr = get_attr (get_conf ("DEFAULTATTR"));
}

void Corpus::set_default_attr (const string &attname)
{
    conf->opts["DEFAULTATTR"] = attname;
    defaultattr = get_attr (attname);
}


Structure* Corpus::get_struct (const string &name)
{
    VSS::iterator i = structs.begin();
    while (i != structs.end()) {
        if ((*i).first == name) 
            return (*i).second;
        i++;
    }
    return setup_struct (name);
}

string Corpus::get_info()
{
    string infostr = get_conf ("INFO");
    if (infostr.empty() || infostr[0] != '@')
        return infostr;
    string s (infostr, 1);
    MapBinFile<char> infofile (get_conf ("PATH") + s);
    s = infofile.at(0);
    return s;
}


const string Corpus::get_sizes()
{
    string sizefile = get_conf ("PATH") + "/sizes";
    ifstream f(sizefile.c_str(), ifstream::in);
    string sizes((istreambuf_iterator<char>(f)),istreambuf_iterator<char>());
    f.close();
    return sizes;
}

const char *manatee_version() {return MANATEE_VERSION_STR;}

// vim: ts=4 sw=4 sta et sts=4 si cindent tw=80:

